;*******************************************************************************
; $RCSfile: qcc-100.as,v $
; $Revision: 1.1 $
; $Author: Scott $
; $Date: 2004/01/26 04:03:58 $
; Description: This file contains the PIC code to convert a quadrature signal
; to a Chinese scale protocol output on the ShumaTech QCC-100.  This version
; features four times resolution multiplication.
;*******************************************************************************
; Copyright (C) 2004 ShumaTech.
; All Rights Reserved.
;*******************************************************************************
; LICENSE
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;*******************************************************************************

; The QCC-100 contains a PIC12F629 but for debugging the software, a PIC16F876
; allows for the use of the in-circuit debugger (ICD).

        ifdef __12F629
        list p=12f629
        errorlevel -302
        include "p12f629.inc"
        endif
        
        ifdef __16F876
        list p=16f876
        errorlevel -302
        include "p16f876.inc"
        endif

;*******************************************************************************
; Configuration Bits
;*******************************************************************************

        ifdef __12F629
        __CONFIG _CPD_OFF & _CP_OFF & _BODEN_ON & _MCLRE_OFF & _PWRTE_ON & _WDT_OFF & _HS_OSC & _EXTRC_OSC_NOCLKOUT
		endif
		
        ifdef __16F876
        __CONFIG _CP_OFF & _DEBUG_ON & _WRT_ENABLE_OFF & _CPD_OFF & _LVP_OFF & _BODEN_ON & _PWRTE_ON & _WDT_OFF & _HS_OSC
        endif
        
;*******************************************************************************
; Constants
;*******************************************************************************
        ifdef __16F876
GPIO		equ	PORTB           ; Use PORTB on the 16F876
TRISIO		equ	TRISB
GPIF		equ	RBIF
GPIE		equ	RBIE
        endif

RAM_START       equ     0x20
RAM_SIZE        equ     0x40

; Three timer periods are required to generate the Chinese scale output.
; The periods are timed with timer1 which is set to run at the internal clock
; rate of Fosc/4 or 5MHz which is equal to a 200ns period.

; The time between number outputs is 20ms.  In timer1 ticks, this is calculated
; as 20ms / 200ns = 100,000.  Since timer1 is only  a 16 bit counter, this time
; is divided into two 10ms time intervals or 50,000 (0xc350) timer1 ticks each.
; Instruction overhead for the delay function call is ignored.
DELAY_10MS_L    equ     0x50
DELAY_10MS_H    equ     0xc3

; The time between the two numbers is 110us and the time before and after the
; number output is 55us.  The number of timer1 ticks is calculated as
; 55us / 200ns = 275.  Subtracting the delay function call overhead
; leaves 250 (0xfa) timer1 ticks.
DELAY_55US      equ     0xfa

; The clock period for the Chinese scale output is 13.2us.  The half period
; between transitions is 6.6us.  The number of timer1 ticks is calculated as
; 6.6us / 200ns = 33.  Subtracting the delay function call overhead
; leaves 10 (0x0a) timer1 ticks.
DELAY_6_6US     equ     0x0a
        
;*******************************************************************************
; Pin Assignments
;*******************************************************************************
        ifdef __12F629
DATA_OUT        equ     0
CLOCK_OUT       equ     1
A_IN            equ     2
B_IN            equ     3
        endif
        
        ifdef __16F876
DATA_OUT        equ     0
CLOCK_OUT       equ     1
A_IN            equ     4
B_IN            equ     2
        endif

;*******************************************************************************
; File Register Variables
;*******************************************************************************
        cblock  RAM_START
w_temp          ; store W during an interrupt
status_temp     ; store STATUS during an interrupt
fsr_temp        ; store FSR during an interrupt

param1          ; parameter 1 for function calls
param2          ; parameter 2 for function calls
temp1           ; temporary scratch variable
temp2           ; temporary scratch variable
temp3           ; temporary scratch variable
        
counter: 3      ; 24 bit quadrature counter
sum             ; sum of products for filter
last_sum        ; last sum of products value
reading: 3      ; 24 bit value to use for the Chines scale output
bits            ; bit iterator
bytes           ; byte iterator
words           ; word iterator
sample_1        ; first GPIO sample
sample_2        ; second GPIO sample
sample_3        ; third GPIO sample
        endc
        
;*******************************************************************************
; Reset Vector
;*******************************************************************************
        org     0x000                   ; processor reset vector
        goto    main                    ; jump to main function

;*******************************************************************************
; Interrupt Service Routine
;
; The ISR looks at the A and B inputs and determines if the counter should be
; incremented, decremented, or left alone due to an input error.  The
; quadrature signals look like the following:
;
;        A Leads B                B Leads A
;      +---+   +---+            +---+   +---+
;      |   |   |   |            |   |   |   |
; A ---+   +---+   +------------+   +---+   +---
;        +---+   +---+        +---+   +---+
;        |   |   |   |        |   |   |   |
; B -----+   +---+   +--------+   +---+   +-----
;      ^ ^ ^ ^ ^ ^ ^          ^ ^ ^ ^ ^ ^ ^ ^
; INC  | | | | | | |      DEC | | | | | | | |
;*******************************************************************************
        org     0x004                   ; processor interrupt vector
        movwf   w_temp                  ; copy W to temp register
        swapf   STATUS,W                ; swap status to be saved into W
        bcf     STATUS,RP0              ; bank 0
        movwf   status_temp             ; save status
        movf    FSR,W                   ; copy FSR to temp register
        movwf   fsr_temp

capture:
        movf    GPIO,W                  ; capture GPIO values
        movwf   sample_1                ; save sample 1
        movf    GPIO,W                  ; capture GPIO values
        movwf   sample_2                ; save sample 2
        movf    GPIO,W                  ; capture GPIO values
        movwf   sample_3                ; save sample 3

        andwf   sample_2,W              ; product term (samples 2 AND 3)
        movwf   sum                     ; save it as the sum

        movf    sample_3,W              ; product term (samples 1 AND 3)
        andwf   sample_1,W              ; 
        iorwf   sum,f                   ; OR it to the sum

        movf    sample_2,W              ; product term (samples 1 AND 2)
        andwf   sample_1,W              ; 
        iorwf   sum,f                   ; OR it to the sum

        movf    sum,W                   ; XOR the last and the current sums
        xorwf   last_sum,F              ; 

        btfss   last_sum,A_IN           ; if the A input has not changed then
        goto    check_b                 ; check if B changed
        btfsc   last_sum,B_IN           ; if the B input has changed too then
        goto    exit_isr                ; exit

check_a:
        btfsc   sum,A_IN                ; if the A input is high then
        goto    rising_a                ; this is a rising edge

falling_a:                              ; else it is a falling edge
        btfsc   sum,B_IN                ; if the B input is high then
        goto    inc_counter             ; increment the counter
        goto    dec_counter             ; else decrement the counter

rising_a:
        btfss   sum,B_IN                ; if the B input is low then
        goto    inc_counter             ; increment the counter
        goto    dec_counter             ; else decrement the counter

check_b:
        btfss   last_sum,B_IN           ; if the B input has not changed then
        goto    exit_isr                ; exit
        btfsc   sum,B_IN                ; if the B input is high then
        goto    rising_b                ; this is a rising edge

falling_b:                              ; else it is a falling edge
        btfss   sum,A_IN                ; if the A input is low then
        goto    inc_counter             ; increment the counter
        goto    dec_counter             ; else decrement the counter

rising_b:
        btfsc   sum,A_IN                ; if the A input is high then
        goto    inc_counter             ; increment the counter
                                        ; else decrement the counter

dec_counter:
        movlw   counter                 ; move the counter address into W
        call    dec24                   ; decrement the counter
        goto    exit_isr
        
inc_counter:
        movlw   counter                 ; move the counter address into W
        call    inc24                   ; increment the counter
                
exit_isr:
        movf    sum,W                   ; save the current sum as the
        movwf   last_sum                ; last
        
        bcf     INTCON,GPIF             ; clear interrupt-on-change flag
                
        movf    fsr_temp,W              ; restore FSR from temp register
        movwf   FSR
        swapf   status_temp,W           ; swap STATUS_TEMP register into W
        movwf   STATUS                  ; move W into STATUS register
        swapf   w_temp,F                ; swap w_temp
        swapf   w_temp,W                ; swap w_temp into W
        retfie                          ; return from the interrupt
        
;*******************************************************************************
; Function:    init
; Description: Intialize the PIC registers
; Parameters:  None
; Returns:     None
;*******************************************************************************
init:
        clrf    GPIO                    ; set all GPIO pins low

        bsf     STATUS,RP0              ; bank 1
        movlw   ~((1 << CLOCK_OUT) | (1 << DATA_OUT))
                                        ; set CLOCK_OUT and DATA_OUT as outputs
        movwf   TRISIO                  ; 1 - input, 0 - output

        ifdef __12F629
        movlw   A_IN                    ; turn on weak pull-up for A_IN
        movwf   WPU                     ;
        movlw   ((1 << A_IN) | (1 << B_IN))
        iorwf   IOC,F                   ; interrupt-on-change for A & B inputs
        endif

        bcf     STATUS,RP0              ; bank 0

        ifdef __12F629
        movlw   0x7                     ; turn off the comparator module
        movwf   CMCON                   ; 111 - comparator off
        endif
        
        bsf     T1CON,TMR1ON            ; enable timer1
        bsf     INTCON,GPIE             ; enable interrupt-on-change interrupt
        bsf     INTCON,GIE              ; enable interrupts
        return
        
;*******************************************************************************
; Function:    delay_cycles
; Description: Delay a specified number of instruction cycles including
;              interrupt cycles.  The function call overhead adds between
;              13 and 16 cycles of delay on top of the specified value.
; Parameters:  param1 - least significant byte of 16 bit cycle delay
;              param2 - most significant byte of 16 bit cycle delay
; Returns:     None
;*******************************************************************************
delay_cycles:
        comf    param1,F                ; negate the delay by complementing the
        comf    param2,F                ; low and high bytes
        bcf     T1CON,TMR1ON            ; stop timer 1
        movf    param1,W                ; move the low byte of the delay into
        movwf   TMR1L                   ; timer 1
        movf    param2,W                ; move the high byte of the delay into
        movwf   TMR1H                   ; timer 1
        bcf     PIR1,TMR1IF             ; clear the timer 1 rollover flag
        bsf     T1CON,TMR1ON            ; turn on timer 1
        
tmr1_check:     
        btfss   PIR1,TMR1IF             ; wait for the timer 1 rollover flag to
        goto    tmr1_check              ; trigger
        return
        
;*******************************************************************************
; Function:    inc24
; Description: Increment a little-endian 24 bit number.  Performance between
;              4 and 10 cycles depending on rollovers.
; Parameters:  W - pointer to the 24 bit number
; Returns:     None
;*******************************************************************************
inc24:  
        movwf   FSR                     ; increment the byte pointed to 
        incfsz  INDF,F                  ; by the indirect register
        return                          ; return if there is no byte rollover

        incfsz  FSR,F                   ; increment the indirect register
        incfsz  INDF,F                  ; increment the byte it points to
        return                          ; return if there is no byte rollover

        incfsz  FSR,F                   ; increment the indirect register
        incf    INDF,F                  ; increment the byte it points to
        return

;*******************************************************************************
; Function:    dec24
; Description: Decrement a little-endian 24 bit number.  Performance between
;              5 and 12 cycles depending on rollovers.
; Parameters:  W - pointer to the 24 bit number
; Returns:     None
;*******************************************************************************
dec24:  
        movwf   FSR                     ; decrement the byte pointed to 
        decfsz  INDF,F                  ; by the indirect register
        incfsz  INDF,W                  ; check the indirect register
        return                          ; and return if it rolled over

        incfsz  FSR,F                   ; decrement the byte pointed to 
        decfsz  INDF,F                  ; by the indirect register
        incfsz  INDF,W                  ; check the indirect register
        return                          ; and return if it rolled over

        incfsz  FSR,F                   ; decrement the byte pointed to 
        decf    INDF,F                  ; by the indirect register
        return
        
;*******************************************************************************
; Main Function
;*******************************************************************************
main:
        call    init                    ; initialize PIC registers

        clrf    counter                 ; zero the counter
        clrf    counter+1               ;
        clrf    counter+2               ;
        
        clrf    reading                 ; zero the reading
        clrf    reading+1               ;
        clrf    reading+2               ;

        clrf    last_sum                ; zero the last sum of samples

start:
        movlw   DELAY_10MS_L            ; Delay for 10 milliseconds
        movwf   param1                  ; 
        movlw   DELAY_10MS_H            ; 
        movwf   param2                  ; 
        call    delay_cycles            ; 

        movlw   DELAY_10MS_L            ; Delay another 10 milliseconds
        movwf   param1                  ; 
        movlw   DELAY_10MS_H            ; 
        movwf   param2                  ; 
        call    delay_cycles            ; 
                
        bcf     INTCON, GIE             ; Disable interrupts
        movf    counter,W               ; Copy the current counter value to
        movwf   reading                 ; an intermediate variable.  Interrupts
        movf    counter+1,W             ; are disabled to protect against the
        movwf   reading+1               ; ISR modifying the counter at the same
        movf    counter+2,W             ; time.
        movwf   reading+2               ; 
        bsf     INTCON,GIE              ; Enable interrupts

        movlw   2                       ; Set the word iterator to 2
        movwf   words                   ;
        
next_word:
        movlw   reading                 ; Load the file register pointer
        movwf   FSR                     ; with the address of the reading
        
        movlw   3                       ; Set the byte iterator to 3
        movwf   bytes                   ; 
        
        movlw   7                       ; Set the bit iterator to 7
        movwf   bits                    ;
        
        bsf     GPIO,CLOCK_OUT          ; Rising clock edge

        btfsc   INDF,0                  ; Output data bit 0
        bsf     GPIO,DATA_OUT
        btfss   INDF,0
        bcf     GPIO,DATA_OUT

        rrf     INDF,F                  ; Advance to next data bit
        
        movlw   DELAY_55US              ; Delay 55 microseconds
        movwf   param1                  ; 
        clrf    param2                  ; 
        call    delay_cycles            ;
                
        bcf     GPIO,CLOCK_OUT          ; Falling clock edge
        
        movlw   DELAY_6_6US             ; Delay 6.6 microseconds
        movwf   temp1
delay_1:  
        decfsz  temp1,F
        goto    delay_1

        bsf     GPIO,CLOCK_OUT          ; Rising clock edge

next_byte:
next_bit:
        btfsc   INDF,0                  ; Output next data bit
        bsf     GPIO,DATA_OUT
        btfss   INDF,0
        bcf     GPIO,DATA_OUT

        rrf     INDF,F                  ; Advance to next data bit
        
        movlw   DELAY_6_6US             ; Delay 6.6 microseconds
        movwf   temp1
delay_2:  
        decfsz  temp1,F
        goto    delay_2

        bcf     GPIO,CLOCK_OUT          ; Falling clock edge
        
        movlw   DELAY_6_6US             ; Delay 6.6 microseconds
        movwf   temp1
delay_3:  
        decfsz  temp1,F
        goto    delay_3
        
        bsf     GPIO,CLOCK_OUT          ; Rising clock edge

        decfsz  bits,F                  ; Decrement the bit iterator
        goto    next_bit                ; and continue if not zero

        rrf     INDF,F                  ; Rotate to the first data bit
        incf    FSR,F                   ; Increment the file register pointer
        movlw   8                       ; Set the bit iterator to 8
        movwf   bits                    ; 
        decfsz  bytes,F                 ; Decrement the byte iterator
        goto    next_byte               ; and continue if not zero
                        
        movlw   DELAY_55US              ; Delay 55 microseconds
        movwf   param1                  ; 
        clrf    param2                  ; 
        call    delay_cycles            ;

        decfsz  words,F                 ; Decrement the word iterator
        goto    next_word               ; and continue if not zero
        
        bcf     GPIO,CLOCK_OUT          ; Clear the clock line
        bcf     GPIO,DATA_OUT           ; Clear the data line

        goto    start
                
        end
